import FTLNomadActorSheet from "./base-actor-sheet.mjs"

export default class FTLNomadCreatureSheet extends FTLNomadActorSheet {
  /** @override */
  static DEFAULT_OPTIONS = {
    classes: ["npc"],
    position: {
      width: 860,
      height: 620,
    },
    window: {
      contentClasses: ["npc-content"],
    },
    actions: {
      createArmor: FTLNomadCreatureSheet.#onCreateArmor,
      createWeapon: FTLNomadCreatureSheet.#onCreateWeapon,
      createSkill: FTLNomadCreatureSheet.#onCreateSkill,

    },
  }

  /** @override */
  static PARTS = {
    main: {
      template: "systems/fvtt-ftl-nomad/templates/creature-main.hbs",
    },
    tabs: {
      template: "templates/generic/tab-navigation.hbs",
    },
    skills: {
      template: "systems/fvtt-ftl-nomad/templates/creature-skills.hbs",
    },
    biography: {
      template: "systems/fvtt-ftl-nomad/templates/creature-biography.hbs",
    },
  }

  /** @override */
  tabGroups = {
    sheet: "skills",
  }

  /**
   * Prepare an array of form header tabs.
   * @returns {Record<string, Partial<ApplicationTab>>}
   */
  #getTabs() {
    const tabs = {
      skills: { id: "skills", group: "sheet", icon: "fa-solid fa-shapes", label: "FTLNOMAD.Label.skills" },
      biography: { id: "biography", group: "sheet", icon: "fa-solid fa-book", label: "FTLNOMAD.Label.biography" },
    }
    for (const v of Object.values(tabs)) {
      v.active = this.tabGroups[v.group] === v.id
      v.cssClass = v.active ? "active" : ""
    }
    return tabs
  }

  /** @override */
  async _prepareContext() {
    const context = await super._prepareContext()
    context.tabs = this.#getTabs()

    context.enrichedDescription = await TextEditor.enrichHTML(this.document.system.description, { async: true })
    context.enrichedNotes = await TextEditor.enrichHTML(this.document.system.notes, { async: true })
    
    return context
  }

  /** @override */
  async _preparePartContext(partId, context) {
    const doc = this.document
    switch (partId) {
      case "main":
        break
      case "skills":
        context.tab = context.tabs.skills
        context.skills = doc.itemTypes.skill
        context.skills.sort((a, b) => a.name.localeCompare(b.name))
        context.weapons = doc.itemTypes.weapon
        context.weapons.sort((a, b) => a.name.localeCompare(b.name))
        context.armors = doc.itemTypes.armor
        context.armors.sort((a, b) => a.name.localeCompare(b.name))
        break
      case "biography":
        context.tab = context.tabs.biography
        context.enrichedDescription = await TextEditor.enrichHTML(doc.system.description, { async: true })
        context.enrichedNotes = await TextEditor.enrichHTML(doc.system.notes, { async: true })
        break
    }
    return context
  }

  /**
   * Creates a new attack item directly from the sheet and embeds it into the document.
   * @param {Event} event             The initiating click event.
   * @param {HTMLElement} target      The current target of the event listener.
   */
  static #onCreateWeapon(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("FTLNOMAD.Label.newWeapon"), type: "weapon" }])
  }

  static #onCreateArmor(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("FTLNOMAD.Label.newArmor"), type: "armor" }])
  }

  static #onCreateSkill(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("FTLNOMAD.Label.newSkill"), type: "skill" }])
  }

  /**
   * Handles the roll action triggered by user interaction.
   *
   * @param {PointerEvent} event The event object representing the user interaction.
   * @param {HTMLElement} target The target element that triggered the roll.
   *
   * @returns {Promise<void>} A promise that resolves when the roll action is complete.
   *
   * @throws {Error} Throws an error if the roll type is not recognized.
   *
   * @description This method checks the current mode (edit or not) and determines the type of roll
   * (save, resource, or damage) based on the target element's data attributes. It retrieves the
   * corresponding value from the document's system and performs the roll.
   */
  async _onRoll(event, target) {
    const rollType = $(event.currentTarget).data("roll-type")
    let item
    let li
    // Debug : console.log(">>>>", event, target, rollType)
    // Deprecated : if (this.isEditMode) return
    switch (rollType) {
      case "char":
        let charId = $(event.currentTarget).data("char-id")
        item = foundry.utils.duplicate(this.actor.system.characteristics[charId])
        item.name = game.i18n.localize(`FTLNOMAD.Label.${charId}Long`)
        item.targetScore = item.value * 5
        break
      case "skill":
        li = $(event.currentTarget).parents(".item");
        item = this.actor.items.get(li.data("item-id"));
        break
      case "weapon":
      case "damage":
        li = $(event.currentTarget).parents(".item");
        item = this.actor.items.get(li.data("item-id"));
        item.damageBonus = this.actor.system.damageBonus
        break
      default:
        throw new Error(`Unknown roll type ${rollType}`)
    }
    await this.document.system.roll(rollType, item)
  }

  async _onDrop(event) {
    if (!this.isEditable || !this.isEditMode) return
    const data = TextEditor.getDragEventData(event)

    // Handle different data types
    switch (data.type) {
      case "Item":
        const item = await fromUuid(data.uuid)
        return super._onDropItem(item)
    }
  }

}
